#include "main.h"

#if !defined _WIN32 || defined _WIN64
#error "JUST FOR WINDOWS 32 bit"
#endif	// if defined _WIN32

static_assert (sizeof (int) == sizeof (long), "JUST FOR WINDOWS 32 bit");
static_assert (sizeof (bool) == sizeof (char), "JUST FOR WINDOWS 32 bit");
static_assert (sizeof (bool) == 1u, "JUST FOR WINDOWS 32 bit");
static_assert (sizeof (int) == 4u, "JUST FOR WINDOWS 32 bit");

enginefuncs_t g_engfuncs;
globalvars_t *gpGlobals;

#define GET_STATIC_STRING_LENGTH(string) (sizeof (string) - sizeof ('\0'))

const unsigned char Pointer8Size (sizeof (unsigned char));
const unsigned char Pointer32Size (sizeof (void *));

// Opcodes with encoding information....
const unsigned char WildCard (0x2A);	// Used for signature scanning. ('*')
const unsigned char JumpImmediately8 (0xEB);	// jmp '.' (encoding is imm8)
const unsigned char JumpImmediately8Size (sizeof (JumpImmediately8) + Pointer8Size);
const unsigned char JumpImmediately32 (0xE9);	// jmp '....' (encoding is imm32)
const unsigned char JumpImmediately32Size (sizeof (JumpImmediately32) + Pointer32Size);	// size of the jump-to instruction
const unsigned char Push32 (0x68);	// push '....' (encoding is <imm32>)
const unsigned char Push32Size (sizeof (Push32) + Pointer32Size);	// size of the push instruction
const unsigned char Nop (0x90);	// nop (no extra encoding)
const unsigned char Return (0xC3);	// ret (no extra encoding)
struct DLL_t
{
	HMODULE      pointer;
	unsigned int size;
};

DLL_t gameDLL = {NULL, 0u};

template <typename elementType> inline const bool CompareMemoryWithoutWildCard (const elementType *const startPointer, const elementType *const pattern, const unsigned char length)
{
	// Compare each byte, except wildcards....
	for (unsigned char index (0u); startPointer[index] == pattern[index]; /* Empty */)
		if (++index == length)	// If 'index' reached the end, we know we have a match!
			return true;	// Found a match.

	return false;
}
template <typename elementType> inline const bool CompareMemory (const elementType *const startPointer, const elementType *const pattern, const unsigned char length, const unsigned char wildCard = WildCard)
{
	// Compare each byte, except wildcards....
	for (unsigned char index (0u); pattern[index] == wildCard || startPointer[index] == pattern[index]; /* Empty */)
		if (++index == length)	// If 'index' reached the end, we know we have a match!
			return true;	// Found a match.

	return false;
}
inline void *const FindPatternWithoutWildCard (void *startPointer, const unsigned int searchLength, const unsigned char *const pattern, const unsigned char length, const unsigned char stepSize = sizeof (unsigned char))
{
	// Scan for the signature in memory then return the starting position's address.

	const unsigned char *const endPointer (static_cast <unsigned char *> (startPointer) + searchLength);	// prevent a crash maybe?

	do
	{
		// Compare each byte, except wildcards....
		if (CompareMemoryWithoutWildCard (static_cast <unsigned char *> (startPointer), pattern, length))
			return startPointer;	// Found a match.
	} while ((reinterpret_cast <unsigned char *&> (startPointer) += stepSize/* search memory in an aligned manner, or not */) + length <= endPointer);	// prevent a crash maybe?

	return NULL;
}
inline void *const FindPattern (void *startPointer, const unsigned int searchLength, const unsigned char *const pattern, const unsigned char length, const unsigned char stepSize = sizeof (unsigned char), const unsigned char wildCard = WildCard)
{
	// Scan for the signature in memory then return the starting position's address.

	const unsigned char *const endPointer (static_cast <unsigned char *> (startPointer) + searchLength);	// prevent a crash maybe?

	do
	{
		// Compare each byte, except wildcards....
		if (CompareMemory (static_cast <unsigned char *> (startPointer), pattern, length, wildCard))
			return startPointer;	// Found a match.
	} while ((reinterpret_cast <unsigned char *&> (startPointer) += stepSize/* search memory in an aligned manner, or not */) + length <= endPointer);	// prevent a crash maybe?

	return NULL;
}
inline void *const FindMemoryChunkReference32 (void *const startPointer, const unsigned int searchLength, const void *const address, const unsigned char instruction)
{
	const union /* Unnamed */
	{
		const void    *pointer;
		unsigned char  byte[Pointer32Size];
	} pointerToByte = {address/*, Do not initialize me! */};
	const unsigned char pattern[sizeof (instruction) + Pointer32Size] =
	{
		instruction, pointerToByte.byte[0u], pointerToByte.byte[1u], pointerToByte.byte[2u], pointerToByte.byte[3u]	// instruction 'address'
	};

	return FindPatternWithoutWildCard (startPointer, searchLength, pattern, sizeof (pattern));
}
inline const bool RedirectFunctionCall (void *const callAddress, const void *const newFunctionAddress, const void **const originalFunctionAddress = NULL)
{
	// Redirect UTIL_IsGame() function call to our function....
	unsigned long oldProtectFlags (0ul);

	if (!VirtualProtect (callAddress, Pointer32Size, PAGE_EXECUTE_READWRITE, &oldProtectFlags))
	{
		g_engfuncs.pfnServerPrint ("RedirectFunctionCall(): VirtualProtect() filed!\n");

		return false;
	}

	// Save the original function address at first...
	if (originalFunctionAddress != NULL)
		*originalFunctionAddress = reinterpret_cast <void *> (static_cast <const unsigned int *const> (callAddress)[0u] + reinterpret_cast <const unsigned int> (callAddress) + Pointer32Size);

	reinterpret_cast <unsigned int *> (callAddress)[0u] = reinterpret_cast <const unsigned int> (newFunctionAddress) - reinterpret_cast <const unsigned int> (callAddress) - Pointer32Size;

	return true;
}

static const bool EnableBotPhraseManagerInitializeFailMessage (void)
{
	return true;
}
static const bool EnableBotConsoleVariableRegister (void)
{
	return true;
}
static const bool EnableBotConsoleCommandRegister (void)
{
	return true;
}
static const bool EnableBotConsoleCommandExecute (void)
{
	return true;
}
static const bool EnableValidateMapData (void)
{
	return true;
}
static const bool EnableAddBot (void)
{
	return true;
}
static const bool EnableBotProfileManagerInitialize (void)
{
	return true;
}
typedef void (*ClientPrint_t) (entvars_t *client, int msg_dest, const char *msg_name, const char *param1, const char *param2, const char *param3, const char *param4);
static ClientPrint_t g_originalClientPrintFunctionAddress (NULL);
static void ClientPrint (entvars_t *client, int msg_dest, const char *msg_name, const char *param1, const char *param2, const char *param3, const char *param4)
{
//	if (strcmp (param1, "#Game_radio_location") == 0)
	if (param1[GET_STATIC_STRING_LENGTH ("#Game_radio")] == '_')
	{
		param1 = "#Game_radio";
		param3 = param4;
		param4 = NULL;
	}

	(*g_originalClientPrintFunctionAddress) (client, msg_dest, msg_name, param1, param2, param3, param4);
}

inline const bool FindGameDLL (void)
{
	gameDLL.pointer = GetModuleHandle ("mp.DLL");

	if (gameDLL.pointer == NULL)
	{
		g_engfuncs.pfnServerPrint ("FindGameDLL(): Can't get mp.DLL handle!!!\n");

		return false;
	}

	// Get mp.DLL size....
	{
		MEMORY_BASIC_INFORMATION info;

		if (VirtualQuery (gameDLL.pointer, &info, sizeof (info)) == 0)
		{
			g_engfuncs.pfnServerPrint ("FindGameDLL(): Can't VirtualQuery() for mp.DLL handle!!!\n");

			return false;
		}

		const IMAGE_DOS_HEADER *const dos (static_cast <IMAGE_DOS_HEADER *> (info.AllocationBase));
		const IMAGE_NT_HEADERS *const pe (reinterpret_cast <IMAGE_NT_HEADERS *> (reinterpret_cast <unsigned int> (info.AllocationBase) + dos->e_lfanew));

		gameDLL.size = pe->OptionalHeader.SizeOfImage;
	}

	return true;
}
inline const bool EnableBotInGame (void)
{
	unsigned char signatureBase[] =
	{
		Push32, WildCard, WildCard, WildCard, WildCard,		// push    offset aCzero   ; "czero"
		0xE8, WildCard, WildCard, WildCard, WildCard,		// call    UTIL_IsGame     ; Call Procedure
		0x83, 0xC4, 0x04,									// add     esp, 4          ; Add
		0x84, 0xC0,											// test    al, al          ; Logical Compare
		0x0F, 0x84, WildCard, WildCard, WildCard, WildCard	// jz      locret_100219AA ; Jump if Zero (ZF=1)
	};	// 21
	const unsigned char signatureBaseCallAddressPosition (Push32Size + sizeof ('\xE8'));

	{
		const unsigned char string[] = {"czero"};
		void *const address (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, GET_STATIC_STRING_LENGTH (string)));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		reinterpret_cast <void **> (signatureBase + sizeof (Push32))[0u] = address;
	}

	// Bot phrase manager initialize fail message:
	{
		const unsigned char string[] = {"WARNING: Cannot access bot phrase database '%s'\n"};
		unsigned char *address (static_cast <unsigned char *> (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, GET_STATIC_STRING_LENGTH (string))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		address = static_cast <unsigned char *> (FindMemoryChunkReference32 (gameDLL.pointer, gameDLL.size, address, Push32));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" push address!!!\n", string));

			return false;
		}

		const unsigned char offset (sizeof (signatureBase) + sizeof ('\x56'));	// 22

		address -= offset;
		address += signatureBaseCallAddressPosition;

		if (!RedirectFunctionCall (address, &EnableBotPhraseManagerInitializeFailMessage))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Bot console variable register:
	{
		const unsigned char signatureLocal[] =
		{
			Push32, WildCard, WildCard, WildCard, WildCard,		// push    offset off_10112BB8 ; _DWORD
			0xFF, 0x15, WildCard, WildCard, WildCard, WildCard,	// call    dword_10162420  ; Indirect Call Near Procedure
			Push32, WildCard, WildCard, WildCard, WildCard,		// push    offset off_10112BD0 ; _DWORD
			0xFF, 0x15, WildCard, WildCard, WildCard, WildCard	// call    dword_10162420  ; Indirect Call Near Procedure
		};
		unsigned char signatureFull[sizeof (signatureBase) + sizeof (signatureLocal)];

		memcpy (signatureFull, signatureBase, sizeof (signatureBase));
		memcpy (signatureFull + sizeof (signatureBase), signatureLocal, sizeof (signatureLocal));

		unsigned char *address (static_cast <unsigned char *> (FindPattern (gameDLL.pointer, gameDLL.size, signatureFull, sizeof (signatureFull))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint ("EnableBotInGame(): Can't find bot console variable register signature address!!!\n");

			return false;
		}

		address += signatureBaseCallAddressPosition;

		if (!RedirectFunctionCall (address, &EnableBotConsoleVariableRegister))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Bot console command register:
	{
		const unsigned char signatureLocal[] =
		{
			0xC6, 0x05, WildCard, WildCard, WildCard, WildCard, 0x00	// mov     byte_1011322C, 0
		};
		unsigned char signatureFull[sizeof (signatureBase) + sizeof (signatureLocal)];	// Just a bit changed 'signatureBase'....

		memcpy (signatureFull, signatureBase, Push32Size);
		memcpy (signatureFull + Push32Size, signatureLocal, sizeof (signatureLocal));
		memcpy (signatureFull + Push32Size + sizeof (signatureLocal), signatureBase + Push32Size, sizeof (signatureBase) - Push32Size);

		unsigned char *address (static_cast <unsigned char *> (FindPattern (gameDLL.pointer, gameDLL.size, signatureFull, sizeof (signatureFull))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint ("EnableBotInGame(): Can't find bot console command register signature address!!!\n");

			return false;
		}

		address += signatureBaseCallAddressPosition + sizeof (signatureLocal);

		if (!RedirectFunctionCall (address, &EnableBotConsoleCommandRegister))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Bot console command execute:
	{
		const unsigned char signatureLocal[] =
		{
			0x89, 0x6C, 0x24, 0x1C	// mov     [esp+1ECh+var_1D0], ebp
		};
		unsigned char signatureFull[sizeof (signatureBase) + sizeof (signatureLocal)];	// Just a bit changed 'signatureBase'....

		memcpy (signatureFull, signatureBase, Push32Size);
		memcpy (signatureFull + Push32Size, signatureLocal, sizeof (signatureLocal));
		memcpy (signatureFull + Push32Size + sizeof (signatureLocal), signatureBase + Push32Size, sizeof (signatureBase) - Push32Size);

		unsigned char *address (static_cast <unsigned char *> (FindPattern (gameDLL.pointer, gameDLL.size, signatureFull, sizeof (signatureFull))));

		if (address == NULL)
		{
			// So try to find signature in new game.DLL...
			const unsigned char signatureFullNewLengthWithoutBase (18u);
			unsigned char signatureFullNew[signatureFullNewLengthWithoutBase + sizeof (signatureBase)] =
			{
				signatureLocal[0u], signatureLocal[1u], signatureLocal[2u], signatureLocal[3u] - 0x04,	// mov [esp+1E8h+var_1D0], ebp
				0x8A, 0x85, 0xE2, 0x02, 0x00, 0x00,														// mov     al, [ebp+2E2h]
				0x84, 0xC0,																				// test    al, al          ; Logical Compare
				0x0F, 0x84, WildCard, WildCard, WildCard, WildCard										// jz      loc_10024859    ; Jump if Zero (ZF=1)
			};

			memcpy (signatureFullNew + signatureFullNewLengthWithoutBase, signatureBase, sizeof (signatureBase));

			// (Address: 0x1002390C)
			address = static_cast <unsigned char *> (FindPattern (gameDLL.pointer, gameDLL.size, signatureFullNew, sizeof (signatureFullNew)));

			if (address == NULL)
			{
				g_engfuncs.pfnServerPrint ("EnableBotInGame(): Can't find bot console command execute signature address!!!");

				return false;
			}

			// (Address: 0x1002391E)
			address += signatureFullNewLengthWithoutBase + signatureBaseCallAddressPosition;
		}
		else
			address += signatureBaseCallAddressPosition + sizeof (signatureLocal);

		if (!RedirectFunctionCall (address, &EnableBotConsoleCommandExecute))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Validate map data:
	{
		const unsigned char string[] = {"Failed to load navigation map.\n"};
		unsigned char *address (static_cast <unsigned char *> (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, GET_STATIC_STRING_LENGTH (string))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		address = static_cast <unsigned char *> (FindMemoryChunkReference32 (gameDLL.pointer, gameDLL.size, address, Push32));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" push address!!!\n", string));

			return false;
		}

		const unsigned char offset (sizeof (signatureBase) + GET_STATIC_STRING_LENGTH ("\xC6\x05\x82\x8F\x12\x10\x01\xE8\x29\x2E\x02\x00\x85\xC0\x74\x15"));	// 37

		address -= offset;
		address += signatureBaseCallAddressPosition;

		if (!RedirectFunctionCall (address, &EnableValidateMapData))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Add bot:
	{
		const unsigned char signatureLocal[] =
		{
			0x8B, 0x74, 0x24, 0x58,									// mov     esi, [esp+50h+arg_4]
			0x8B, 0x3D, WildCard, WildCard, WildCard, WildCard,		// mov     edi, dword_10162304
			0x83, 0xFE, 0x02										// cmp     esi, 2          ; Compare Two Operands
		};
		unsigned char signatureFull[sizeof (signatureBase) + sizeof (signatureLocal)];

		memcpy (signatureFull, signatureBase, sizeof (signatureBase));
		memcpy (signatureFull + sizeof (signatureBase), signatureLocal, sizeof (signatureLocal));

		unsigned char *address (static_cast <unsigned char *> (FindPattern (gameDLL.pointer, gameDLL.size, signatureFull, sizeof (signatureFull))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint ("EnableBotInGame(): Can't find bot console variable register signature address!!!\n");

			return false;
		}

		address += signatureBaseCallAddressPosition;

		if (!RedirectFunctionCall (address, &EnableAddBot))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	// Bot profile manager initialize:
	{
		const unsigned char string[] = {"WARNING: Cannot access bot profile database '%s'\n"};
		unsigned char *address (static_cast <unsigned char *> (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, GET_STATIC_STRING_LENGTH (string))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		address = static_cast <unsigned char *> (FindMemoryChunkReference32 (gameDLL.pointer, gameDLL.size, address, Push32));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" push address!!!\n", string));

			return false;
		}

		const unsigned char offset (sizeof (signatureBase) + sizeof ('\x57'));	// 22

		address -= offset;
		address += signatureBaseCallAddressPosition;

		if (!RedirectFunctionCall (address, &EnableBotProfileManagerInitialize))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't disable UTIL_IsGame() function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}
//#if 0
	// Remove "-nobots" launch parameter check:
	int metaMajor (0), metaMinor (0);

	sscanf (g_engfuncs.pfnCVarGetString ("metamod_version"), "%i.%i", &metaMajor, &metaMinor);

	if (metaMajor <= 1 && metaMinor < 21)
	{
		g_engfuncs.pfnServerPrint ("EnableBotInGame(): You use old version of metamod! Trying to fix \"-nobots\" crash... (Please update metamod)\n");

		const unsigned char string[] = {"-nobots"};
		unsigned char *address (static_cast <unsigned char *> (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, GET_STATIC_STRING_LENGTH (string))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		address = static_cast <unsigned char *> (FindMemoryChunkReference32 (gameDLL.pointer, gameDLL.size, address, Push32));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" push address!!!\n", string));

			return false;
		}

		address -= GET_STATIC_STRING_LENGTH ("\x6A\x00");
/*ALSO WORKS
		const unsigned char length (GET_STATIC_STRING_LENGTH ("\x6A\x00\x68\x90\x38\x11\x10\xFF\x15\xB4\x25\x16\x10\x83\xC4\x08\x85\xC0"));	// 18
		unsigned long oldProtectFlags (0ul);

		if (!VirtualProtect (address, length + sizeof (JumpImmediately8), PAGE_EXECUTE_READWRITE, &oldProtectFlags))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't VirtualProtect() for address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}

		for (unsigned char index (0u); index < length; ++index, ++address)
			address[0u] = Nop;

		address[0u] = JumpImmediately8;	// Replace 'jz' to 'jmp'.
*/
		// Construct jump trough UTIL_IsGame() function call....
		unsigned long oldProtectFlags (0ul);

		if (!VirtualProtect (address, JumpImmediately8Size, PAGE_EXECUTE_READWRITE, &oldProtectFlags))
		{
			g_engfuncs.pfnServerPrint ("RemoveUTIL_IsGameFunctionCall(): VirtualProtect() filed!\n");

			return false;
		}

		address[0u] = JumpImmediately8;

		unsigned char jumpEndIndex
		(
			GET_STATIC_STRING_LENGTH
			(
				"\x6A\x00"					// push    0               ; _DWORD
				"\x68\x90\x38\x11\x10"		// push    offset aNobots  ; "-nobots"
				"\xFF\x15\xB4\x25\x16\x10"	// call    dword_101625B4  ; Indirect Call Near Procedure
				"\x83\xC4\x08"				// add     esp, 8          ; Add
				"\x85\xC0"					// test    eax, eax        ; Logical Compare
				"\x74\x08"					// jz      short loc_10026212 ; Jump if Zero (ZF=1)
			)
		);

		while (address[++jumpEndIndex] != Return)
		{
			// Reliability check.
			if (jumpEndIndex == 0u)
			{
				g_engfuncs.pfnServerPrint ("RemoveUTIL_IsGameFunctionCall(): Can't find 'retn' opcode!\n");

				return false;
			}
		}

		++jumpEndIndex;	// Skip 'Return'

		reinterpret_cast <unsigned int *> (address + sizeof (JumpImmediately8))[0u] = jumpEndIndex - JumpImmediately8Size;
	}
//#endif	// if 0
	// Remove location from radio messages:
	{
		const unsigned char string[] = {"#Game_radio"};
		unsigned char *address (static_cast <unsigned char *> (FindPatternWithoutWildCard (gameDLL.pointer, gameDLL.size, string, sizeof (string))));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" address!!!\n", string));

			return false;
		}

		address = static_cast <unsigned char *> (FindMemoryChunkReference32 (gameDLL.pointer, gameDLL.size, address, Push32));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't find string \"%s\" push address!!!\n", string));

			return false;
		}

		const unsigned char signatureLocal[] =
		{
			0xE8, WildCard, WildCard, WildCard, WildCard,	// call    sub_100CE480    ; Call Procedure
			0x83, 0xC4, 0x1C								// add     esp, 1Ch        ; Add
		};

		address = static_cast <unsigned char *> (FindPattern (address, 30u, signatureLocal, sizeof (signatureLocal)));

		if (address == NULL)
		{
			g_engfuncs.pfnServerPrint ("EnableBotInGame(): Can't find 'ClientPrint' function call address!!!\n");

			return false;
		}

		++address;	// Skip '0xE8' (call) opcode.

		if (!RedirectFunctionCall (address, &ClientPrint, reinterpret_cast <const void **const> (&g_originalClientPrintFunctionAddress)))
		{
			g_engfuncs.pfnServerPrint (va ("EnableBotInGame(): Can't redirect 'ClientPrint' function call on address 0x%X!!!\n", address - (unsigned char *)gameDLL.pointer));

			return false;
		}
	}

	return true;
}

void ToggleAdvancedHostageAI (void)
{
	static unsigned char originalBytes[GET_STATIC_STRING_LENGTH ("\xA1\x50\xA8\x12\x10")];
	const unsigned char replacedBytes[] = {0xB8, 0x00, 0x00, 0x00, 0x00};
	unsigned char *const address (reinterpret_cast <unsigned char *> (GetProcAddress (gameDLL.pointer, "?IdleThink@CHostage@@QAEXXZ")));

	if (address == NULL)
	{
		g_engfuncs.pfnServerPrint ("ToggleAdvancedHostageAI(): FILED!!! Can't find CHostage::IdleThink() function!\n");

		return;
	}

	if (address[0u] == 0xA1)
	{
		// Not patched yet.

		if (address[1u] != 0x50 || address[sizeof (originalBytes)] != 0x83 || address[sizeof (originalBytes) + 1u] != 0xEC)
		{
			g_engfuncs.pfnServerPrint ("ToggleAdvancedHostageAI(): FILED!!! Can't find pattern!\n");

			return;
		}

		memcpy (originalBytes, address, sizeof (originalBytes));

		unsigned long oldProtectFlags (0ul);

		if (!VirtualProtect (address, sizeof (replacedBytes), PAGE_EXECUTE_READWRITE, &oldProtectFlags))
		{
			g_engfuncs.pfnServerPrint ("ToggleAdvancedHostageAI(): VirtualProtect() filed!\n");

			return;
		}

		memcpy (address, replacedBytes, sizeof (replacedBytes));
	}
	else
	{
		// Patched already.

		memcpy (address, originalBytes, sizeof (originalBytes));
	}
}

#if _MSC_VER > 1000
#pragma comment(linker, "/EXPORT:GiveFnptrsToDll=_GiveFnptrsToDll@8,@1")
#pragma comment(linker, "/SECTION:.data,RW")
#endif

C_DLLEXPORT void WINAPI GiveFnptrsToDll(enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   memcpy(&g_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));
   gpGlobals = pGlobals;

	if (!FindGameDLL ())
	{
		g_engfuncs.pfnServerPrint ("FindGameDLL(): FILED!!!\n");

		return;
	}

	if (EnableBotInGame ())
	{
		g_engfuncs.pfnAddServerCommand ("csbotenable_toggle_advanced_hostage_AI", &ToggleAdvancedHostageAI);

		g_engfuncs.pfnServerPrint ("EnableBotInGame(): SUCCEED!!!\n");
	}
}

static META_FUNCTIONS gMetaFunctionTable = {
   GetEntityAPI, NULL, NULL, NULL, NULL, NULL, GetEngineFunctions, NULL,
};

plugin_info_t Plugin_info = {
	META_INTERFACE_VERSION,
	"Enable CSBot",
	__DATE__,
	__DATE__,
	"Whistler <whistler_wmz@users.sf.net>",
	"http://yapb.bots-united.com",
	"BOTENAB",
	PT_ANYTIME,
	PT_ANYPAUSE,
};

meta_globals_t *gpMetaGlobals;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;

C_DLLEXPORT int Meta_Query(char *ifvers, plugin_info_t **pPlugInfo,
    mutil_funcs_t *pMetaUtilFuncs) {
   *pPlugInfo = &Plugin_info;
   gpMetaUtilFuncs = pMetaUtilFuncs;
   return TRUE;
}

C_DLLEXPORT int Meta_Attach(PLUG_LOADTIME now,
    META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, 
    gamedll_funcs_t *pGamedllFuncs) {
   gpMetaGlobals = pMGlobals;

   memcpy(pFunctionTable, &gMetaFunctionTable, sizeof(META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;
   return TRUE;
}

C_DLLEXPORT int Meta_Detach(PLUG_LOADTIME, PL_UNLOAD_REASON) {
   return TRUE;
}

